/*
	physdiskwrite.c - program to write binary image files to raw disks (CF cards)

	Copyright (C) 2002 Manuel Kasper <mk@neon1.net>.
	All rights reserved.
	
	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:
	
	1. Redistributions of source code must retain the above copyright notice,
	   this list of conditions and the following disclaimer.
	
	2. Redistributions in binary form must reproduce the above copyright
	   notice, this list of conditions and the following disclaimer in the
	   documentation and/or other materials provided with the distribution.
	
	THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
	AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
	AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
	OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <windows.h>
#include <winioctl.h>

/* maximum allowed disk size in MB */
#define DISK_MAX_SIZE 1638400	/* 800 MB */

#define READ_BUF_SIZE 8192
#define PHYSDISKWRITE_VERSION "0.1"

void usage(void) {

	printf("Usage: physdiskwrite <image-file>\n");
}

void main(int argc, char *argv[]) {

	int	i = 0, diskid;
	char tmp[128], diskok[32];
	HANDLE hdisk, himg;
	DISK_GEOMETRY	dge;
	DWORD	br, bw, n, imgsize;

	printf("\nphysdiskwrite v" PHYSDISKWRITE_VERSION " by Manuel Kasper <mk@neon1.net>\n\n");

	if (argc != 2) {
		usage();
		return;
	}

	printf("Searching for physical drives...\n\n");

	/* search for physical drives
	   at the moment, it seems like the only option is to keep trying devices
	   until CreateFile fails */

	while (i < 32) {
		sprintf(tmp, "\\\\.\\PhysicalDrive%d", i);

		/* try to open it */
		hdisk = CreateFile(tmp,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL);

		if (hdisk == INVALID_HANDLE_VALUE)
			break;

		/* try to obtain the disk geometry. It looks like this function returns
		   incorrect values for CF cards. With real hard drives, the values seem
		   to be correct. */
		if (DeviceIoControl(hdisk,
							IOCTL_DISK_GET_DRIVE_GEOMETRY,
							NULL,
							0,
							(LPVOID)&dge,
							sizeof(dge),
							&br,
							NULL)) {

			printf("Geometry for %s:\n", tmp);
			printf("   cyl: %lu\n   tpc: %lu\n   spt: %lu\n   bps: %lu\n\n",
				dge.Cylinders.LowPart, dge.TracksPerCylinder,
				dge.SectorsPerTrack, dge.BytesPerSector);

			/* safety check - if the disk is larger than DISK_MAX_SIZE,
			   we will record this fact in diskok[] and prevent writing the disk later
			   as it is probably not a CF card
			*/
			if ((dge.Cylinders.LowPart * dge.TracksPerCylinder * dge.SectorsPerTrack)
				> DISK_MAX_SIZE) {
				diskok[i] = 0;
			} else {
				diskok[i] = 1;
			}

		} else {
			printf("DeviceIoControl() failed on %s.\n", tmp);
		}

		CloseHandle(hdisk);
		i++;
	}

	do {
		printf("Which disk do you want to write? (0..%d) ", i-1);
		scanf("%d", &diskid);
		fflush(stdin);

		if ((diskid < 0) || diskid > (i-1))
			continue;

		if (!diskok[diskid]) {
			printf("That disk is larger than 800 MB (safety overwrite check). Please choose another one.\n");
		}
	} while (!diskok[diskid]);

	printf("About to overwrite the contents of disk %d with new data. Proceed? (y/n) ", diskid);

	if (getchar() == 'y') {

		char	*buf;

		/* allocate a buffer */
		buf = VirtualAlloc(NULL, READ_BUF_SIZE, MEM_COMMIT, PAGE_READWRITE);

		if (buf == NULL) {
			fprintf(stderr, "Failed to allocate buffer - exiting.\n");
			return;
		}

		sprintf(tmp, "\\\\.\\PhysicalDrive%d", diskid);

		hdisk = CreateFile(tmp,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL);

		if (hdisk == INVALID_HANDLE_VALUE) {
			fprintf(stderr, "Failed to open %s - exiting.\n", tmp);
			return;
		}

		himg = CreateFile(argv[1],
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL);

		if (hdisk == INVALID_HANDLE_VALUE) {
			fprintf(stderr, "Failed to open %s - exiting.\n", argv[1]);
			return;
		}

		imgsize = GetFileSize(himg, NULL);

		n = 0;
		while (ReadFile(himg, buf, READ_BUF_SIZE, &br, NULL)) {
			if (br) {
				if (!WriteFile(hdisk, buf, READ_BUF_SIZE, &bw, NULL)) {
					fprintf(stderr, "Write error after %lu bytes.\n", n);
					return;
				}
				n += br;
				printf("\r%lu/%lu bytes written", n, imgsize);
			} else {
				break;
			}
		}
		printf("\r%lu/%lu bytes written in total\n", n, imgsize);

		CloseHandle(himg);
		CloseHandle(hdisk);

		VirtualFree(buf, 0, MEM_RELEASE);
	}
}
